/// Josh's User Interface Library
/// A C++20 Library for creating, styling, and rendering of a UI/UX widgets.
/// Developed and Maintained by Josh O'Leary @ Redacted Software.
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file Button.hpp
/// @desc Clickable Button widget.
/// @edit 2024-07-31

#pragma once

#include <JUI/Base/Widget.hpp>
#include <JUI/Base/RectBase.hpp>
#include <JUI/Base/TextBase.hpp>
#include <JUI/Mixins/Clickable.hpp>
#include <JUI/Mixins/Hoverable.hpp>

namespace JUI
{
    enum class ButtonState
    {
        Base,
        Hovered,
        Pressed,
        Disabled
    };

    /// The Button class is a basic rect widget that accepts mouse inputs, and provides event hooks.
    /// The button also provides built-in automatic coloring based on it's current state, which are listed above.
    class Button : public Rect, public Clickable, public Hoverable
    {
    public:
        Event<> OnEnabled;
        Event<> OnDisabled;
    public:
        Button();
        explicit Button(Widget* parent);
        ~Button() override {};
    public:

        /// Returns the background color of this object when the mouse is hovering this button.
        /// @note Setting BGColor and BorderColor directly on this class is not recommended, as it will be overwritten.
        [[nodiscard]] Color4 HoveredBGColor() const;
        /// Returns the background color of this object at rest.
        /// @note Setting BGColor and BorderColor directly on this class is not recommended, as it will be overwritten.
        [[nodiscard]] Color4 BaseBGColor() const;
        /// Returns the background color of this object when it is clicked.
        /// @note Setting BGColor and BorderColor directly on this class is not recommended, as it will be overwritten.
        [[nodiscard]] Color4 PressedBGColor() const;
        /// Returns the background color of this object when it is disabled.
        /// @note Setting BGColor and BorderColor directly on this class is not recommended, as it will be overwritten.
        [[nodiscard]] Color4 DisabledBGColor() const;
        /// Returns the border color of this object when it is hovered.
        /// @note Setting BGColor and BorderColor directly on this class is not recommended, as it will be overwritten.
        [[nodiscard]] Color4 HoveredBorderColor() const;
        /// Returns the border color of this object at rest.
        /// @note Setting BGColor and BorderColor directly on this class is not recommended, as it will be overwritten.
        [[nodiscard]] Color4 BaseBorderColor() const;
        /// Returns the border color of this object when it is clicked.
        /// @note Setting BGColor and BorderColor directly on this class is not recommended, as it will be overwritten.
        [[nodiscard]] Color4 PressedBorderColor() const;
        /// Returns the border color of this object when it is disabled.
        /// @note Setting BGColor and BorderColor directly on this class is not recommended, as it will be overwritten.
        [[nodiscard]] Color4 DisabledBorderColor() const;

        /// Sets the background color of this object when the mouse is hovering.
        void HoveredBGColor(const Color4& color);
        /// Sets the background color of this object at rest.
        void BaseBGColor(const Color4& color);
        /// Sets the background color of this object when it is clicked.
        void PressedBGColor(const Color4& color);
        /// Sets the background color of this object when it is disabled.
        void DisabledBGColor(const Color4& color);

        /// Sets the border color of this object when the mouse is hovering.
        void HoveredBorderColor(const Color4& color);
        /// Sets the border color of this object at rest.
        void BaseBorderColor(const Color4& color);
        /// Sets the border color of this object when it is clicked.
        void PressedBorderColor(const Color4& color);
        /// Sets the border color of this object when it is disabled.
        void DisabledBorderColor(const Color4& color);

        void BGColors(const Color4& base, const Color4& hover, const Color4& pressed, const Color4& disabled = Colors::Gray);

        void BorderColors(const Color4& base, const Color4& hover, const Color4& pressed, const Color4& disabled = Colors::Gray);

        /// Returns whether is button is interactable. If enabled, it will listen to mouse events and react accordingly.
        bool Enabled() const;
        /// Returns whether is button is interactable. If enabled, it will listen to mouse events and react accordingly.
        bool Disabled() const;
        /// Sets the interactable state of this button.
        void SetEnabled(bool enabled);

        /// Disables this button. Mouse events will be ignored and the button will grey-out.
        void Disable();
        /// Enables this button. @see Disable, SetEnabled
        void Enable();

        void Update(float delta) override;

        void OnClick(const Vector2& mouse_pos, const MouseButton& btn) override;
        void OnRelease(const Vector2& mouse_pos, const MouseButton& bnt, bool still_hovering) override;
        void OnHover(const J3ML::LinearAlgebra::Vector2 &MousePos) override;
        void OnExit(const J3ML::LinearAlgebra::Vector2 &MousePos) override;
        //void SetTooltip(const std::string &content, float delay = 0.2f) override;
    protected:
        bool disabled = false;

        Color4 hover_bg           = Style::Button::HoverBackgroundColor;
        Color4 hover_border       = Style::Button::HoverBorderColor;
        Color4 pressed_bg         = Style::Button::PressBackgroundColor;
        Color4 pressed_border     = Style::Button::PressBorderColor;
        Color4 disabled_bg        = Style::Button::DisableBackgroundColor;
        Color4 disabled_border    = Style::Button::DisableBorderColor;
        Color4 base_bg            = Style::Button::BaseBackgroundColor;
        Color4 base_border        = Style::Button::BaseBorderColor;

        void UpdateVisualState();
    };


}